# pyright: strict, reportUnusedFunction=false, reportUnknownVariableType=false
import queue
from typing import Any, AsyncGenerator, Generator, Callable, Optional, Dict, List
from typing_extensions import TypedDict
from concurrent.futures import ThreadPoolExecutor
import json
import time
import os
import asyncio
from ..utils.log import *
from .tokenizer import TokenizerPool
import inspect
import threading
import traceback
import uvicorn
from fastapi import FastAPI, Request
from fastapi.responses import StreamingResponse
from contextlib import asynccontextmanager
from abc import ABC, abstractmethod
from ..utils.util import APP_ID

ENGINE_READY : bool = False

SERVER_CODE_OK = 0
SERVER_CODE_ERROR = 1
SERVER_CODE_STOP = 2

_APP_NAME = os.getenv("APP_NAME", "ieops")

_MODEL_SERVER_SOCKET = os.getenv("MODEL_SERVER_SOCKET", f"{APP_ID}.sock")
_MODEL_SERVER_PORT = os.getenv("MODEL_SERVER_PORT", 8001)
_MODEL_SERVER_HOST = os.getenv("MODEL_SERVER_HOST", "127.0.0.1")

_THREAD_CONCURRENCY = os.getenv("MODEL_THREAD_CONCURRENCY", "8")
_MODEL_GRPC_PORT = os.getenv("MODEL_GRPC_PORT", 8001)
_MODEL_TIMEOUT = os.getenv("MODEL_TIMEOUT", 600)
_GRACEFUL_SHUTDOWN_TIME = os.getenv("GRACEFUL_SHUTDOWN_TIME", 3)
_MAX_MESSAGE_LENGTH = os.getenv("MAX_MESSAGE_LENGTH", 10*1024*1024)


class ModelError(Exception):
    def __init__(self, message: str):
        super().__init__(message)
        self.message = message


class QueueItem(TypedDict):
    code: int
    data: bytes


class Server(ABC):
    @abstractmethod
    def shutdown(self) -> None:
        """request server graceful shutdown"""
    @abstractmethod
    async def serve(self) -> None:
        """serve the server"""


class OpenAIServer(Server):
    def __init__(self, 
    func: Callable[[Any], Generator[Any, bool, None]], 
    abort : Optional[Callable[[str], None]] = None) -> None:
        self._func = func
        self._cancelled_events : Dict[str,bool] = {}
        self._abort = abort
        self._executor = ThreadPoolExecutor(max_workers = int(_THREAD_CONCURRENCY) * 2 + 3)
        self._app=FastAPI(lifespan=self._lifespan)
        self._server: Optional[uvicorn.Server] = None
        self._setup_routes()

    @asynccontextmanager
    async def _lifespan(self, app: FastAPI):
        # start
        uvicorn_logger.info("Server starting...")
        yield
        # clean up when shutting down
        uvicorn_logger.info("Server shutting down...")
        # cancel all in-progress requests
        for trace_id in list(self._cancelled_events.keys()):
            self._cancelled_events[trace_id] = True
        # give in-progress requests a chance to complete (optional)
        await asyncio.sleep(0.5)  # wait for 0.5 seconds to let requests have a chance to check the cancel flag
        self._executor.shutdown(wait=True)
        try:
            TokenizerPool(None).shutdown(wait=True)
        except Exception as e:
            uvicorn_logger.error(f"Error shutting down TokenizerPool: {e}")
        finally:
            if _MODEL_SERVER_SOCKET and os.path.exists(_MODEL_SERVER_SOCKET):
                os.remove(_MODEL_SERVER_SOCKET)
    
    def shutdown(self) -> None:
        """request server graceful shutdown"""
        if self._server:
            self._server.should_exit = True

    async def serve(self) -> None:
        # configure uvicorn
        if _MODEL_SERVER_SOCKET:
            config = uvicorn.Config(
                self._app,
                uds=_MODEL_SERVER_SOCKET,
                log_level="info",
                loop="asyncio",
                timeout_graceful_shutdown=int(_GRACEFUL_SHUTDOWN_TIME)  # graceful shutdown timeout
            )
        else:
            config = uvicorn.Config(
                self._app,
                host=_MODEL_SERVER_HOST,
                port=int(_MODEL_SERVER_PORT),
                log_level="info",
                loop="asyncio",
                timeout_graceful_shutdown=int(_GRACEFUL_SHUTDOWN_TIME)  # graceful shutdown timeout
            )
        
        self._server = uvicorn.Server(config)
        await self._server.serve()

    def _setup_routes(self) -> None:
        """set API endpoints"""
        @self._app.get("/")
        async def root():
            """Root endpoint with basic info"""
            return {
                "message": "vLLM Load Balancing Server",
                "status": "ready" if ENGINE_READY else "initializing",
                "endpoints": {
                    "health": "/ping",
                    "generate": "/v1/completions",
                    "chat": "/v1/chat/completions"
                }
            }
        @self._app.get("/health")
        async def health():
            return {"status": "ok"}
        
        @self._app.post("/sse/infer")
        async def infer_sse(request: Request):
            body = await request.body()
            return StreamingResponse(
                self._sse_generator(body),
                media_type="text/event-stream",
                headers={
                    "Cache-Control": "no-cache",
                    "Connection": "keep-alive",
                    "X-Accel-Buffering": "no",
                }
            )
        
        @self._app.post("/infer")
        async def infer(request: Request):
            """non-streaming interface, return complete JSON response"""
            body = await request.body()
            results: List[Any] = []
            last_code = SERVER_CODE_OK
            async for item in self._generator(body):
                code = item["code"]
                data = item["data"]
                if isinstance(data, bytes):
                    data = data.decode("utf-8")
                if code == SERVER_CODE_ERROR:
                    last_code = code
                    results.append({"error": data})
                    break
                if code == SERVER_CODE_STOP:
                    break
                try:
                    results.append(json.loads(data))
                except json.JSONDecodeError:
                    results.append(data)
            return {"code": last_code, "results": results}

    async def _sse_generator(self, request: bytes) -> AsyncGenerator[str, None]:
        """SSE response generator"""
        async for item in self._generator(request):
            code = item["code"]
            data = item["data"]
            if isinstance(data, bytes):
                data = data.decode("utf-8")
            yield f"{json.dumps({'code': code, 'data': data}, ensure_ascii=False)}\n\n"

    async def _generator(self, request: bytes) -> AsyncGenerator[Dict[str, Any], None]: 
        start = int(time.time() * 1000)
        loop = asyncio.get_event_loop()
        headers: Dict[str, str] = {}
        payload: Any = None
        try:
            request_data = json.loads(request.decode("utf-8"))
            headers = request_data.get("headers", {})
            payload = request_data.get("payload", request_data)  # compatible with old format
        except json.JSONDecodeError:
            # 如果不是 JSON，当作纯 payload 处理
            payload = request.decode("utf-8")
            
        trace_id: str = str(headers.get("x-trace-id", "traceid-" + await arandstr(8)))
        log.get_logger(trace_id=trace_id).info("query size: {}", len(payload) if isinstance(payload, (str, bytes)) else len(json.dumps(payload)))
        # 如果 payload 已经是 dict，直接使用；否则解析 JSON
        query: Dict[str, Any] = payload if isinstance(payload, dict) else json.loads(payload)
        query['trace_id'] = trace_id
        if trace_id in self._cancelled_events:
            log.get_logger(trace_id=trace_id).warning("trace_id already exists")
            yield {"code": SERVER_CODE_ERROR, "data": f"{trace_id} already exists".encode("utf-8")}
            return
        self._cancelled_events[trace_id] = False
        all_data : List[Any] = []
        try:
            if inspect.isasyncgenfunction(self._func):
                threading.Thread(target=self._time_counter, args=(trace_id, int(_MODEL_TIMEOUT))).start()
                async for v in self._func(query):
                    if self._cancelled_events[trace_id]:
                        del self._cancelled_events[trace_id]
                        if self._abort is not None:
                            await self._abort(trace_id) # type: ignore
                        raise asyncio.TimeoutError
                    if v is None:
                        break
                    if isinstance(v, ModelError):
                        yield {"code": SERVER_CODE_ERROR, "data": v.message.encode("utf-8")}
                        all_data.append(" [ model error: "+v.message+" ]")
                        break
                    if "results" in v:
                        for ret in v["results"]:
                            all_data.append(ret["content"])
                    yield {"code": SERVER_CODE_OK, "data": json.dumps(v, ensure_ascii=False).encode("utf-8")}
            else:
                q : "queue.Queue[QueueItem]" = queue.Queue()
                # await loop.run_in_executor(self._executor, self._cancelled_events[token].clear)
                await asyncio.wait_for(loop.run_in_executor(self._executor, self._func_caller, trace_id, query, q), int(_MODEL_TIMEOUT)//10) # type: ignore
                while True:
                    v:Any = await loop.run_in_executor(self._executor, q.get)
                    if v['code'] == SERVER_CODE_STOP:
                        yield {"code": SERVER_CODE_STOP, "data": b""}
                        break
                    if v['code'] == SERVER_CODE_ERROR:
                        yield {"code": SERVER_CODE_ERROR, "data": json.dumps(v, ensure_ascii=False).encode("utf-8")}
                        break
                    yield {"code": SERVER_CODE_OK, "data": json.dumps(v, ensure_ascii=False).encode("utf-8")}
        except asyncio.CancelledError as e:
            log.get_logger(trace_id=trace_id).error("cancelled: {}", e)
            yield {"code": SERVER_CODE_ERROR, "data": b"cancelled"} 
        except asyncio.TimeoutError as e:
            log.get_logger(trace_id=trace_id).error("timeout: {}", e)
            yield {"code": SERVER_CODE_ERROR, "data": b"timeout"}
        except Exception as e:
            log.get_logger(trace_id=trace_id).error("error: {}\n{}", e, traceback.format_exc())
            try:
                yield {"code": SERVER_CODE_ERROR, "data": traceback.format_exc().encode("utf-8")}
            except Exception as e:
                log.get_logger(trace_id=trace_id).error("error: {}\n{}", e, traceback.format_exc().encode("utf-8"))
        finally:
            log.get_logger(trace_id=trace_id).info("all responses length: {}, cost: {}ms", len(all_data), int(time.time() * 1000) - start)
            # await loop.run_in_executor(self._executor, self._cancelled_events[token].set)
            if trace_id in self._cancelled_events:
                self._cancelled_events[trace_id] = True
            log.get_logger(trace_id=trace_id).info("all content: {}", "".join(all_data))

    def _func_caller(self, trace_id: str, payload: Dict[str, Any], q: "queue.Queue[QueueItem]"):
        try:
            it = self._func(payload)
            if not hasattr(it, "send"):
                # 不是生成器
                q.put(QueueItem(code=SERVER_CODE_OK, data=json.dumps(it, ensure_ascii=False).encode("utf-8")))
                q.put(QueueItem(code=SERVER_CODE_STOP, data=b""))
                return
            v = it.send(None) # pyright: ignore
            while True:
                q.put(QueueItem(code=SERVER_CODE_OK, data=json.dumps(v, ensure_ascii=False).encode("utf-8")))
                if self._cancelled_events[trace_id]:
                    try:
                        v = it.send(True)
                    except StopIteration:
                        break
                    break
                else:
                    try:
                        v = it.send(False)
                    except StopIteration:
                        break
        except Exception:
            # TODO: log exception here
            log.get_logger(trace_id=trace_id).error("error: {}", traceback.format_exc())
            q.put(QueueItem(code=SERVER_CODE_ERROR, data=traceback.format_exc().encode("utf-8")))
        else:
            # put none to stop the loop
            q.put(QueueItem(code=SERVER_CODE_STOP, data=b""))
        finally:
            log.get_logger(trace_id=trace_id).info("func_caller finished")
            del self._cancelled_events[trace_id]

    def _time_counter(self, trace_id : str, timeout: int):
        for _ in range(timeout):
            if self._cancelled_events[trace_id]:
                log.get_logger(trace_id=trace_id).info("time counter stop") # type: ignore
                del self._cancelled_events[trace_id]
                return
            time.sleep(0.1)
        self._cancelled_events[trace_id] = True


class ComfyUIServer(Server):
    def __init__(self) -> None:
        raise NotImplementedError("ComfyUIServer is not implemented")

    def shutdown(self) -> None:
        """request server graceful shutdown"""
        pass

    async def serve(self) -> None:
        """serve the server"""
        pass


__all__ = ["Server", "OpenAIServer", "ComfyUIServer", "ENGINE_READY", "ModelError"]


import os
from ..utils.util import randstr, APP_ID
from ..utils.log import *
import aiohttp
import asyncio
import requests
from typing import Dict, Any, TypedDict, Optional

_APP_NAME = os.getenv("APP_NAME", "ieops")
_SERVER_SOCKET_DIR = os.getenv("SERVER_SOCKET_DIR", os.getcwd()) # 默认是项目根目录

_IEOPS_PROXY_URL = os.getenv("IEOPS_PROXY_URL", "http://10.146.0.6:30054/api/model")
_MODEL_THREAD_CONCURRENCY = os.getenv("MODEL_THREAD_CONCURRENCY", "8")
_REGISTER_INTERVAL = os.getenv("REGISTER_INTERVAL", 10)
_REGISTER_ENABLED = os.getenv("REGISTER_ENABLED", "true")

# async connector (lazy init)
_CONNECTOR: Optional[aiohttp.TCPConnector] = None

# sync session (lazy init)
_SYNC_SESSION: Optional[requests.Session] = None


class Payload(TypedDict):
    path: str
    payload: Dict[str, Any]


def _get_async_connector() -> aiohttp.TCPConnector:
    """lazy init async connector (needs to be called in event loop)"""
    global _CONNECTOR
    if _CONNECTOR is None:
        _CONNECTOR = aiohttp.TCPConnector()
    return _CONNECTOR


def _get_sync_session() -> requests.Session:
    global _SYNC_SESSION
    if _SYNC_SESSION is None:
        _SYNC_SESSION = requests.Session()
    return _SYNC_SESSION


def send(_payload: Payload) -> Optional[Exception]:
    """send request to IEOPS proxy"""
    try:
        session = _get_sync_session()
        url = f"{_IEOPS_PROXY_URL}/{_payload['path']}"
        resp = session.post(url, json=_payload['payload'], timeout=0.8)
        if resp.status_code == 200:
            uvicorn_logger.info(f"Client send request to [{_payload['path']}] success")
            return None
        return Exception(f"Client send request to [{_payload['path']}] failed: {resp.status_code}")
    except requests.Timeout as e:
        uvicorn_logger.warning(f"Client send request to [{_payload['path']}] timeout: {e}")
        return e
    except requests.RequestException as e:
        uvicorn_logger.warning(f"Client send request to [{_payload['path']}] failed: {e}")
        return e


async def async_send(_payload: Payload) -> Optional[Exception]:
    """async send request to IEOPS proxy"""
    err = None
    try:
        url = f"{_IEOPS_PROXY_URL}/{_payload['path']}"
        client_timeout = aiohttp.ClientTimeout(total=0.8)
        async with aiohttp.ClientSession(
            connector=_get_async_connector(), 
            connector_owner=False,
            timeout=client_timeout
        ) as session:
            async with session.post(url, json=_payload['payload']) as resp:
                if resp.status == 200:
                    # uvicorn_logger.info(f"Client send request to [{_payload['path']}] success")
                    return None
                err = Exception(f"Client send request to [{_payload['path']}] failed: {resp.status}")
    except asyncio.TimeoutError:
        uvicorn_logger.warning(f"Client send request to [{_payload['path']}] timeout")
        err = Exception(f"Request timeout after 0.8s")
    except aiohttp.ClientError as e:
        uvicorn_logger.warning(f"Client send request to [{_payload['path']}] failed: {e}")
        err = e
    except asyncio.CancelledError:
        uvicorn_logger.info("Client stopped")
        raise
    return err


class Register:
    def __init__(self):
        self._appinfo = self._app_info()
        
    def _app_info(self):
        self._tokens = [randstr(8) for _ in range(int(_MODEL_THREAD_CONCURRENCY))]
        return {
            "id": APP_ID,
            "server_socket": f"{_SERVER_SOCKET_DIR}/{APP_ID}.sock",
            "max_concurrent_reqs": int(_MODEL_THREAD_CONCURRENCY),
            "endpoint": _APP_NAME,
            "weight": 1,
        }

    async def register(self):
        if _REGISTER_ENABLED.lower() == "false":
            uvicorn_logger.info("Register is disabled, skipping...")
            return
        uvicorn_logger.info("Registering {} to IEOPS proxy...".format(_APP_NAME))
        try:
            while True:
                await async_send(Payload(path="register", payload=self._appinfo))
                await asyncio.sleep(int(_REGISTER_INTERVAL))
        except asyncio.CancelledError:
            uvicorn_logger.info("Register stopped")
            raise
        finally:
            if _CONNECTOR is not None:
                await _CONNECTOR.close()
            if _SYNC_SESSION is not None:
                _SYNC_SESSION.close()

__all__ = ['Payload', 'send', 'async_send', 'Register']


CLOSE = "QUEUE.CLOSE"


class Node(object):

    def __init__(self, value=None, next_node=None):
        self.value = value
        self.next = next_node

    def set_value(self, value):
        self.value = value

    def set_next(self, next_node):
        self.next = next_node


class LinkQueue(object):

    def __init__(self):
        self.front = Node()  # 头节点
        self.rear = self.front  # 队列为空时，头节点=尾节点
        self.count = 0

    def __len__(self):
        return self.count

    def __str__(self):
        s = ''
        cursor = self.front
        while cursor != self.rear:
            cursor = cursor.next
            s += ' %s' % cursor.value
        return s

    def empty(self):
        return self.count == 0

    def put(self, value):
        node = Node(value)
        self.rear.next = node
        self.rear = node
        self.count += 1
        return True

    def close(self):
        node = Node(CLOSE)
        self.rear.next = node
        self.rear = node
        self.count += 1
        return True

    def get(self):
        if self.front == self.rear:
            return None
        p = self.front.next
        value = p.value
        self.front.next = p.next
        # 若队头为等于队尾，则尾节点指向头节点
        if self.rear == p:
            self.rear = self.front
        del p
        self.count -= 1
        return value



# pyright: strict, reportUnusedFunction=false

from concurrent.futures import ProcessPoolExecutor
from typing import List, Dict, Union, Optional, Any, TYPE_CHECKING
import asyncio
import threading
import multiprocessing
from abc import ABC, abstractmethod
import signal
from ..utils.log import log, uvicorn_logger

if TYPE_CHECKING:
    from transformers import BatchEncoding  # type: ignore
    import numpy as np
    import torch  # type: ignore

class TokenizerBase(ABC):
    """Tokenizer abstract base class"""
    @abstractmethod
    def init_tokenizer(self) -> None:
        """initialize tokenizer"""
        pass

    @abstractmethod
    def encode(self, msg: Union[str, List[Dict[str, str]], List[List[Dict[str, str]]]]) -> Union[str, List[int], List[str], List[List[int]], Any]:
        """encode interface"""
        pass
    @abstractmethod
    def decode(self, token_ids: Union[int, List[int], Any]) -> str:
        """decode interface"""
        pass

class TokenizerPool():
    _instance = None
    _lock = threading.Lock()
    _initialized = False
    _shutting_down = False
    _process_pool: Optional[ProcessPoolExecutor] = None

    def __new__(cls, *args, **kwargs): # type: ignore
        with cls._lock:
            if cls._instance is None:
                cls._instance = super().__new__(cls)
            return cls._instance

    def __init__(
        self,
        tokenizer: Union[TokenizerBase, None],
        max_workers: int = multiprocessing.cpu_count()
    ):
        if not self._initialized:
            with self._lock:
                if not self._initialized:
                    if tokenizer is None:
                        raise ValueError("Tokenizer must be provided")
                    self._tokenizer = tokenizer
                    self._max_workers = max_workers
                    
                    # initialize tokenizer in main process first
                    self._tokenizer.init_tokenizer()
                    
                    # create process pool
                    self._start_pool()
                    
                    self._loop = asyncio.get_event_loop()
                    self._initialized = True

    def _start_pool(self):
        """start process pool"""
        try:
            self._process_pool = ProcessPoolExecutor(
                max_workers=self._max_workers,
                initializer=self._init_worker,
                initargs=(self._tokenizer,)
            )
            self._process_pool.submit(self._init_worker, self._tokenizer)
            uvicorn_logger.info(f"Started tokenizer pool with {self._max_workers} workers")
        except Exception as e:
            log.get_logger().error(f"Failed to start process pool: {e}")
            raise

    @staticmethod
    def _init_worker(tokenizer: TokenizerBase):
        """initialize worker process"""
        # set signal handler for child process
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        signal.signal(signal.SIGTERM, signal.SIG_IGN)
        
        # initialize tokenizer
        global _worker_tokenizer
        _worker_tokenizer = tokenizer
        _worker_tokenizer.init_tokenizer()

    @staticmethod
    def _encode_worker(msg: Union[List[Dict[str, str]], List[List[Dict[str, str]]]], tokenize: bool = True) -> "Union[str, List[int], List[str], List[List[int]], BatchEncoding]": # type: ignore
        """worker process encode function"""
        global _worker_tokenizer
        if not hasattr(_worker_tokenizer, 'tokenizer') or _worker_tokenizer.tokenizer is None: # type: ignore
            raise RuntimeError("Worker tokenizer not initialized")
        return _worker_tokenizer.encode(msg, tokenize) # type: ignore

    @staticmethod
    def _decode_worker(token_ids: Union[int, List[int], "np.ndarray", "torch.Tensor"]) -> str: # type: ignore
        """worker process decode function"""
        global _worker_tokenizer
        if not hasattr(_worker_tokenizer, 'tokenizer') or _worker_tokenizer.tokenizer is None: # type: ignore
            raise RuntimeError("Worker tokenizer not initialized")
        return _worker_tokenizer.decode(token_ids) # type: ignore

    async def encode(self, msg: Union[List[Dict[str, str]], List[List[Dict[str, str]]]], tokenize: bool = True) -> "Union[str, List[int], List[str], List[List[int]], BatchEncoding]": # type: ignore
        """async encode interface"""
        if self._shutting_down:
            raise RuntimeError("TokenizerPool is shutting down")
        if not self._initialized or not self._process_pool:
            raise RuntimeError("TokenizerPool not initialized")
        try:
            return await self._loop.run_in_executor(self._process_pool, self._encode_worker, msg, tokenize) # type: ignore
        except Exception as e:
            log.get_logger().error(f"Encode error: {e}")
            raise

    async def decode(self, token_ids: Union[int, List[int], "np.ndarray", "torch.Tensor"]) -> str: # type: ignore
        """async decode interface"""
        if self._shutting_down:
            raise RuntimeError("TokenizerPool is shutting down")
        if not self._initialized or not self._process_pool:
            raise RuntimeError("TokenizerPool not initialized")
        try:
            return await self._loop.run_in_executor(self._process_pool, self._decode_worker, token_ids) # type: ignore
        except Exception as e:
            log.get_logger().error(f"Decode error: {e}")
            raise

    def shutdown(self, wait: bool = True) -> None:
        """shutdown process pool"""
        if not hasattr(self, '_initialized') or not self._initialized or self._shutting_down:
            return
        with self._lock:
            if not self._initialized or self._shutting_down:
                return
            self._shutting_down = True
            try:
                if self._process_pool:
                    uvicorn_logger.info("Shutting down TokenizerPool...")
                    # save process reference
                    # processes = {}
                    # if hasattr(self._process_pool, '_processes'):
                    #     processes = self._process_pool._processes.copy() # type: ignore
                    # cancel all pending tasks
                    if hasattr(self._process_pool, '_work_queue'):
                        try:
                            self._process_pool._work_queue.queue.clear()  # type: ignore
                        except Exception as e:
                            log.get_logger().warning(f"Error clearing work queue: {e}")
                    # shutdown process pool
                    self._process_pool.shutdown(wait=False)
                    # wait for processes to finish
                    if wait and self._process_pool._processes:
                        for pid, p in self._process_pool._processes.items():
                            try:
                                p.kill()
                            except Exception as e:
                                log.get_logger().warning(f"Error terminating process {pid}: {e}")
                                
            except Exception as e:
                log.get_logger().error(f"Error during shutdown TokenizerPool: {e}")
            finally:
                if self._initialized:
                    uvicorn_logger.info("TokenizerPool shutdown completed")
                self._process_pool = None
                self._initialized = False
                self._shutting_down = False

    def __del__(self):
        """destructor"""
        if not self._shutting_down:
            self.shutdown(wait=False)


import random
import os

_TOKEN_CHARS = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"


async def arandstr(n: int):
    return ''.join(random.choices(_TOKEN_CHARS, k=n))


async def arandint(a, b):
    return int(random.random()*(b-a))+a


def randstr(n: int):
    return ''.join(random.choices(_TOKEN_CHARS, k=n))


def randint(a, b):
    return int(random.random()*(b-a))+a

APP_ID = f"{os.getenv('APP_NAME', 'ieops')}.{randstr(8)}"



from setuptools import setup, find_packages
from pathlib import Path

# 读取 requirements.txt
requirements = Path(__file__).parent.joinpath('requirements.txt').read_text().splitlines()

sdk_packages = find_packages(where='src')
packages = ['gmi_ieops'] + [f'gmi_ieops.{p}' for p in sdk_packages]

setup(
    name="gmi_ieops",
    version="0.1.1",
    author="GMICloud Inc.",
    packages=packages,
    package_dir={'gmi_ieops': 'src'},
    python_requires=">=3.10",
    install_requires=requirements
)



import os
import json
from typing import Dict, Any, Union, Literal
from ..handler.client import *
from ..utils.log import *

# 支持的数据类型
FileData = Union[Dict[str, Any], bytes, str]
# write mode
WriteMode = Literal["w", "a"]  # write mode: w=overwrite, a=append


def load_json(file_path: str) -> Dict[str, Any]:
    with open(file_path, 'r') as f:
        return json.load(f)


def save_json(file_path: str, data: Dict[str, Any]):
    with open(file_path, 'w') as f:
        json.dump(data, f)


def _write_file(file_path: str, data: FileData, mode: WriteMode = "w") -> None:
    """
    automatically choose write mode based on data type
    
    :param file_path: file path
    :param data: data (bytes/str/dict)
    :param mode: write mode: w=overwrite, a=append
    """
    if isinstance(data, bytes):
        # binary data (image, audio, etc.)
        binary_mode = f"{mode}b"  # binary mode: wb or ab
        with open(file_path, binary_mode) as f:
            f.write(data)
    elif isinstance(data, str):
        # text data
        with open(file_path, mode, encoding='utf-8') as f:
            f.write(data)
    elif isinstance(data, dict):
        # JSON data
        if mode == "a":
            # append mode: write JSON (JSONL format)
            with open(file_path, 'a', encoding='utf-8') as f:
                f.write(json.dumps(data, ensure_ascii=False) + '\n')
        else:
            # overwrite mode
            with open(file_path, 'w', encoding='utf-8') as f:
                json.dump(data, f, ensure_ascii=False)
    else:
        raise TypeError(f"Unsupported data type: {type(data)}")


def save(file_path: str, data: FileData, mode: WriteMode = "w") -> Optional[Exception]:
    """
    save file to JFS and sync
    
    :param file_path: file path
    :param data: data (bytes/str/dict)
    :param mode: write mode: w=overwrite, a=append
    """
    try:
        _write_file(file_path, data, mode)
        return None
    except Exception as e:
        return e


async def async_save_jfs(file_path: str, data: FileData, mode: WriteMode = "w") -> Optional[Exception]:
    """
    async save file to JFS and sync
    
    :param file_path: file path
    :param data: data (bytes/str/dict)
    :param mode: write mode: w=overwrite, a=append
    """
    try:
        _write_file(file_path, data, mode)
        return await async_send(Payload(path="jfs_sync", payload={"file_path": file_path}))
    except Exception as e:
        return e


def save_jfs(file_path: str, data: FileData, mode: WriteMode = "w") -> Optional[Exception]:
    """
    save file to JFS and sync
    
    :param file_path: file path
    :param data: data (bytes/str/dict)
    :param mode: write mode: w=overwrite, a=append
    """
    try:
        _write_file(file_path, data, mode)
        return send(Payload(path="jfs_sync", payload={"file_path": file_path}))
    except Exception as e:
        return e


from .handler import *
from .tokenizer import *
from .server import *
from .client import *
from ..utils.util import randstr
import os

_APP_ID = f"{os.getenv('APP_NAME', 'ieops')}.{randstr(8)}"
__all__ = ['Handler', 'TokenizerPool', 'TokenizerBase', 'OpenAIServer', 'Client', '_APP_ID']


from .queue import *
from .util import *
from .llm_sampling_params import *
from .log import *
from .util import *
from .file import *

log.set_logger(
    log_path=os.getenv("LOG_FILE_PATH", "/var/log/ieops"),
    app_name=os.getenv("APP_NAME", "ieops"),
    log_level=os.getenv("LOG_LEVEL", "INFO")
)   